/* Minor tweaks to the ReadTheDocs theme */

/* Slightly wider content area */
.wy-nav-content {
    max-width: 900px;
}

/* Monospace font for inline code in parameter tables */
.rst-content dl dt code {
    font-size: 90%;
}

/* Tighten up parameter description lists */
.rst-content dl.field-list dd {
    margin-bottom: 0.4em;
}
